<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "h2devLib.h"
#include "posterLib.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "private.h"
#include "serialize.h"


/* --- local data ---------------------------------------------------------- */

<'foreach s [$component services] {'>
int	genom_<"$comp">_<"[$s name]">_encode(char *buffer, int size,
		char *dst, int maxsize);
int	genom_<"$comp">_<"[$s name]">_decode(char *buffer, int size,
		char *dst, int maxsize);
<'}'>
<'foreach p [$component ports out] {'>
int	genom_<"$comp">_<"[$p name]">_decode(char *buffer,
		<"[[$p datatype] argument reference dst]">);
<'}'>

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set noack [catch {$s task}]'>
  int rid;
  int s;
  assert(h);

  h->error = pocolibs_newrqst(
    h, sizeof(struct genom_<"$comp">_<"[$s name]">_output), rqstid);
  if (h->error) return h->error;

  h->error = csClientRqstSend(
    h->csid, <"$COMP">_<"[$s name]">_RQSTID,
    (char *)in, 0, genom_<"$comp">_<"[$s name]">_encode,
    <"[expr ${noack} ? "FALSE" : "TRUE"]">, 0, 0, &rid);
  if (h->error != OK) return h->error = errnoGet();

  genom_<"$comp">_client_<"[$s name]">_init_output(
    h->active_rqst[*rqstid].output);
  h->active_rqst[*rqstid].info =
    genom_<"$comp">_client_<"[$s name]">_info;
  h->active_rqst[*rqstid].rid = rid;
  h->active_rqst[*rqstid].decode = genom_<"$comp">_<"[$s name]">_decode;
  h->active_rqst[*rqstid].sentcb = (genom_request_cb)sentcb;
  h->active_rqst[*rqstid].donecb = (genom_request_cb)donecb;
  h->active_rqst[*rqstid].cb_data = cb_data;
<'  if {$noack} {'>
  h->active_rqst[*rqstid].status = WAITING_FINAL_REPLY;
<'  } else {'>
  h->active_rqst[*rqstid].status = WAITING_INTERMED_REPLY;

  if (!sentcb) {
    s = pocolibs_waitack(h, *rqstid);
    if (s) return s;
  }
<'  }'>

  if (!donecb) {
    s = pocolibs_waitterm(h, *rqstid);
    if (s) return s;
  }

  return 0;
}

<'}'>

<'foreach p [$component ports out simple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  <"[[$p datatype] argument reference data]">)
{
  static POSTER_ID id;
  int searched = 0;
  char *addr;
  int s;

  do {
    if (!id) {
      char name[H2_DEV_MAX_NAME];
      int n;

      n = snprintf(name, sizeof(name), "%s/<"[$p name]">", h->instance);
      if (n <= 0 || n >= sizeof(name)) return h->error = ENAMETOOLONG;

      if (posterFind(name, &id) != OK) return h->error = errnoGet();
      searched = 1;
    }

    if (posterTake(id, POSTER_READ) != OK) {
      h->error = errnoGet();
      if (!searched && h->error == S_posterLib_POSTER_CLOSED) {
        id = NULL; continue /* try to refetch poster id */;
      }
      return h->error;
    }
    addr = posterAddr(id);
    s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
    posterGive(id);
    break;
  } while(1);

  return s;
}

<'}'>
<'foreach p [$component ports out multiple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  const char *path,
  <"[[$p datatype] argument reference data]">)
{
  char name[H2_DEV_MAX_NAME];
  POSTER_ID id;
  char *addr;
  int s, n;

  n = snprintf(name, sizeof(name), "%s/<"[$p name]">/%s", h->instance, path);
  if (n <= 0 || n >= sizeof(name)) return h->error = ENAMETOOLONG;

  if (posterFind(name, &id) != OK) return h->error = errnoGet();
  if (posterTake(id, POSTER_READ) != OK) return h->error = errnoGet();

  addr = posterAddr(id);
  s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
  posterGive(id);
  return s;
}

<'}'>
