<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Mon Apr  9 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

set types [pubtypes no]

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#ifdef __linux__
# if defined(HAVE_VASPRINTF) && defined(HAVE_ASPRINTF)
/* Apparently we need _GNU_SOURCE defined to get access to vasprintf on Linux */
#  ifndef _GNU_SOURCE
#   define _GNU_SOURCE
#  endif
# endif
#endif

#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <math.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "<"$comp">/c/client.h"


/* --- local data ---------------------------------------------------------- */

static int	bufcat(char **buf, char *fmt, ...);

<'foreach t $types {'>
static __inline__ int	json_scan_<"[mangle $t]">(
  <"[$t argument reference]">, const char **);
static __inline__ int	json_print_<"[mangle $t]">(char **,
  <"[$t argument value]">);
<'}'>

#define json_skip_whitespace(json) (json) += strspn(json, " \t\n\r")

<'foreach s [$component services] {'>

/* <"[--- Service [$s name] encoding/decoding ---------------------------]"> */

int
genom_<"$comp">_client_<"[$s name]">_json_scan(
  struct genom_<"$comp">_<"[$s name]">_input *in, const char *json,
  char **endptr)
{
  int s = 0;
  json_skip_whitespace(json);
  if (*(json++) != '{') { s = EINVAL; goto done; }
  json_skip_whitespace(json);
  while(*json != '}') {
    if (*(json++) != '"') { s = EINVAL; goto done; };
<'    foreach p [$s parameters in inout] {'>
    if (!strncmp(json, "<"[$p name]">\"", 1+<"[string length [$p name]]">)) {
      json += 1+<"[string length [$p name]]">;
      json_skip_whitespace(json);
      if (*(json++) != ':') { s = EINVAL; goto done; }
      if ((s = json_scan_<"[mangle [$p type]]">(
             <"[[$p type] pass reference in->[cname [$p name]]]">, &json)))
        goto done;
    } else
<'    }'>
    { s = EINVAL; goto done; }

    json_skip_whitespace(json);
    if (*json == ',') {
      json++;
      json_skip_whitespace(json);
    }
  }

done:
  if (endptr) *endptr = (char *)json;
  return s;
}

int
genom_<"$comp">_client_<"[$s name]">_json_print(char **json,
             const struct genom_<"$comp">_<"[$s name]">_output *out)
{
  int s = 0;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) free(*json);
    *json = NULL;
    return s;
  }

<'  set sep {}'>
<'  foreach p [$s parameters inout out] {'>
  if ((s = bufcat(json, "<"$sep">\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }
  if ((s = json_print_<"[mangle [$p type]]">(json,
         <"[[$p type] pass value out->[$p name]]">))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }
<'  set sep ,'>
<'    }'>

  if ((s = bufcat(json, "}"))) {
    if (*json) free(*json);
    *json = NULL;
    return s;
  }

  return s;
}

<'}'>

<'foreach p [$component port out] {'>

/* <"[--- Port [$p name] encoding ---------------------------------------]"> */

int
genom_<"$comp">_client_<"[$p name]">_json_print(char **json,
  <"[[$p datatype] argument value data]">)
{
  int s;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }

  if ((s = bufcat(json, "\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }
  if ((s = json_print_<"[mangle [$p datatype]]">(json, data))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }

  if ((s = bufcat(json, "}"))) {
    if (*json) { free(*json); *json = NULL; }
    return s;
  }

  return 0;
}

<'}'>

<'foreach t $types {'>

/* <"[--- [mangle $t] ===================================================]"> */

static __inline__ int
json_scan_<"[mangle $t]">(<"[$t argument reference data]">, const char **json)
{
  int s = 0;
<'
  switch -glob -- [$t kind] {
    typedef - {* member} {'>
  if ((s = json_scan_<"[mangle [$t type]]">(data, json))) return s;
<'
    }
    struct - union {'>
  json_skip_whitespace(*json);
  if (*((*json)++) != '{') return EINVAL;
  json_skip_whitespace(*json);
  while(**json != '}') {
    if (*((*json)++) != '"') return EINVAL;
<'      foreach e [$t members] {'>
    if (!strncmp(*json, "<"[$e name]">\"", 1+<"[string length [$e name]]">)) {
      (*json)+= 1+<"[string length [$e name]]">;
      json_skip_whitespace(*json);
      if (*((*json)++) != ':') return EINVAL;
      if ((s = json_scan_<"[mangle [$e type]]">(
             <"[$e pass reference data->[cname [$e name]]]">, json))) return s;
    } else
<'      }'>
      return EINVAL;

    json_skip_whitespace(*json);
    if (**json == ',') {
      (*json)++;
      json_skip_whitespace(*json);
    }
  }
  (*json)++;
<'
    }
    sequence {'>
  long i;

  json_skip_whitespace(*json);
  switch (*((*json)++)) {
    case '[':
      json_skip_whitespace(*json);
      i = 0;
      while(**json != ']') {
<'      if {[catch {$t length} l]} {'>
        if ((s = genom_sequence_reserve(data, i+1))) return s;
<'      } else {'>
        if (i >= <"$l">) return EINVAL;
<'      }'>
        if ((s = json_scan_<"[mangle [$t type]]">(
               <"[[$t type] pass reference data->_buffer\[i\]]">, json)))
          return s;
        data->_length = ++i;
        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        }
        i++;
      }
      break;

    case '{':
      json_skip_whitespace(*json);
      while(**json != '}') {
        char *d;

        if (*((*json)++) != '"') return EINVAL;
        i = strtol(*json, &d, 0);
        if (d == *json || *d != '"' || i<0) return EINVAL;
        *json = d+1;
        json_skip_whitespace(*json);
        if (*((*json)++) != ':') return EINVAL;
<'      if {[catch {$t length} l]} {'>
        if (data->_length < i+1) {
          if ((s = genom_sequence_reserve(data, i+1))) return s;
        }
<'      } else {'>
        if (i >= <"$l">) return EINVAL;
<'      }'>
        if ((s = json_scan_<"[mangle [$t type]]">(
               <"[[$t type] pass reference data->_buffer\[i\]]">, json)))
          return s;
        if (data->_length < i+1) data->_length = i+1;
        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        }
      }
      break;

    default: return EINVAL;
  }
  (*json)++;
<'
    }
    array {'>
  long i;

  json_skip_whitespace(*json);
  switch (*((*json)++)) {
    case '[':
      json_skip_whitespace(*json);
      i = 0;
      while(**json != ']') {
        if (i >= <"[$t length]">) return EINVAL;
        if ((s = json_scan_<"[mangle [$t type]]">(
               <"[[$t type] pass reference data\[i\]]">, json))) return s;
        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        }
        i++;
      }
      break;

    case '{':
      json_skip_whitespace(*json);
      while(**json != '}') {
        char *d;

        if (*((*json)++) != '"') return EINVAL;
        i = strtol(*json, &d, 0);
        if (d == *json || *d != '"' || i<0) return EINVAL;
        *json = d+1;
        json_skip_whitespace(*json);
        if (*((*json)++) != ':') return EINVAL;
        if (i >= <"[$t length]">) return EINVAL;
        if ((s = json_scan_<"[mangle [$t type]]">(
               <"[[$t type] pass reference data\[i\]]">, json))) return s;
        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        }
      }
      break;

    default: return EINVAL;
  }
  (*json)++;
<'
    }
    char {'>
  json_skip_whitespace(*json);
  if (*((*json)++) != '"') return EINVAL;
  *data = *((*json)++);
  if (*((*json)++) != '"') return EINVAL;
<'
    }
    string {'>
  const char *d;

  json_skip_whitespace(*json);
  if (**json != '"') return EINVAL;
  d = ++(*json);
  while (**json != '"') {
    if (**json == '\\') (*json)++;
    (*json)++;
  }
<'      if {[catch {$t length} l]} {'>
  if (*data) free(*data);
  *data = malloc(*json - d + 1);
  if (!*data) return ENOMEM;
  strncpy(*data, d, *json - d);
  (*data)[*json - d] = '\0';
<'      } else {'>
  if (*json - d + 1 > <"$l">) return EINVAL;
  strncpy(data, d, *json - d);
  data[*json - d] = '\0';
<'      }'>
  (*json)++;
<'
    }
    enum {'>
  json_skip_whitespace(*json);
  if (*((*json)++) != '"') return EINVAL;
<'      foreach e [$t members] {'>
  if (!strncmp(*json, "<"[$e fullname]">\"",
               1+<"[string length [$e fullname]]">)) {
    (*json)+= 1+<"[string length [$e fullname]]">;
    *data = <"[cname [$e fullname]]">;
  } else
    if (!strncmp(*json, "<"[$e name]">\"", 1+<"[string length [$e name]]">)) {
    (*json)+= 1+<"[string length [$e name]]">;
    *data = <"[cname [$e fullname]]">;
  } else
<'      }'>
    return EINVAL;
<'
    }
    float - double {'>
  double v;

  if (!strncmp(*json, "null", 4)) {
    (*json) += 4;
    *data = (<"[$t declarator]">)nan("");
  } else {
    char *d;
    v = strtod(*json, &d);
    if (d == *json) return EINVAL;
    *json = d;
    *data = (<"[$t declarator]">)v;
  }
<'
    }
    default {'>
  long long v;
  char *d;

  v = strtoll(*json, &d, 0);
  if (d == *json) return EINVAL;
  *json = d;
  *data = (<"[$t declarator]">)v;
<'
    }
  }'>
  return s;
}

static __inline__ int
json_print_<"[mangle $t]">(char **json, <"[$t argument value data]">)
{
  int s;
<'
  switch -glob -- [$t kind] {
    typedef - {* member} {'>
  if ((s = json_print_<"[mangle [$t type]]">(json, data))) return s;
<'
    }
    struct - union {'>
  if ((s = bufcat(json, "{"))) return s;
<'
      set sep {}
      foreach e [$t members] {'>
  if ((s = bufcat(json, "<"$sep">\"<"[$e name]">\":"))) return s;
  if ((s = json_print_<"[mangle $e]">(json,
         <"[$e pass value data->[cname [$e name]]]">))) return s;
<'
        set sep ,
      }'>
  if ((s = bufcat(json, "}"))) return s;
<'
    }
    sequence {'>
  size_t i;
  if ((s = bufcat(json, "{"))) return s;
  for (i=0; i<data->_length; i++) {
    if ((s = bufcat(json, "%s\"%d\":", i?",":"", i))) return s;
    if ((s = json_print_<"[mangle [$t type]]">(
           json, <"[[$t type] pass value {data->_buffer[i]}]">))) return s;
  }
  if ((s = bufcat(json, "}"))) return s;
<'
    }
    array {'>
  size_t i;
  if ((s = bufcat(json, "{"))) return s;
  for (i=0; i<<"[$t length]">; i++) {
    if ((s = bufcat(json, "%s\"%d\":", i?",":"", i))) return s;
    if ((s = json_print_<"[mangle [$t type]]">(
           json, <"[[$t type] pass value {data[i]}]">))) return s;
  }
  if ((s = bufcat(json, "}"))) return s;
<'
    }
    char {'>
  if ((s = bufcat(json, "\"%c\"", *data))) return s;
<'
      }
    string {'>
  if ((s = bufcat(json, "\"%s\"", data))) return s;
<'
    }
    enum {'>
  switch(*data) {
<'      foreach e [$t members] {'>
    case <"[cname [$e fullname]]">:
      if ((s = bufcat(json, "\"<"[$e fullname]">\""))) return s; else break;
<'      }'>
  }
<'
    }
    float - double {'>
  if (isnan(*data)) {
    if ((s = bufcat(json, "null"))) return s;
  } else {
    if ((s = bufcat(json, "%g", *data))) return s;
  }
<'
    }
    {unsigned *} {'>
  uint64_t v = *data;
  if ((s = bufcat(json, "%" PRIu64, v))) return s;
<'
    }
    default {'>
  int64_t v = *data;
  if ((s = bufcat(json, "%" PRIi64, v))) return s;
<'
    }
  }'>
  return 0;
}
<'}'>

/* --- bufcat -------------------------------------------------------------- */

static int
bufcat(char **buf, char *fmt, ...)
{
  va_list ap;
  int s;
#if defined(HAVE_VASPRINTF) && defined(HAVE_ASPRINTF)
  char *r = NULL;

  va_start(ap, fmt);
  s = vasprintf(&r, fmt, ap);
  va_end(ap);
  if (s < 0) return ENOMEM;

  if (*buf) {
    s = asprintf(buf, "%s%s", *buf, r);
    free(r);
    if (s < 0) return ENOMEM;
  } else
    *buf = r;
#else
  static char buf1[1024];

  s = 1 + (*buf ? strlen(*buf) : 0);

  va_start(ap, fmt);
  vsnprintf(buf1, sizeof(buf1), fmt, ap);
  va_end(ap);
  s += strlen(buf1);

  *buf = realloc(*buf, s);
  if (!*buf) return ENOMEM;
  strcat(*buf, buf1);
#endif

  return 0;
}
