#
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Mon Feb 27 2012
#


# --- pubtypes -------------------------------------------------------------
#
# Return the list of all public types (recursive)
#
proc pubtypes { private args } {
  set pub_types [dict create]

  foreach comp [dotgen components] {
    expandtype [$comp event] pub_types

    foreach p [$comp ports] {
      expandtype [$p datatype] pub_types
      if {$private} { expandtype [$p type] pub_types }
    }

    foreach s [$comp services] {
      foreach p [$s parameters] {
        switch [$p dir] {
          "inport" - "outport" {
            # handled above
          }
          default {
            expandtype [$p type] pub_types
          }
        }
      }
    }

    if {$private} {
      foreach r [$comp remotes] {
        foreach p [$r parameters] { expandtype [$p type] pub_types }
      }

      if {![catch {$comp ids} ids]} { expandtype $ids pub_types }
    }
  }

  foreach t $args { expandtype $t pub_types }

  return [dict values $pub_types]
}

proc expandtype { type done& } {
  upvar ${done&} done
  set name [mangle $type]
  if {[dict exists $done $name]} return

  switch -- [$type kind] {
    {array} - {sequence} - {typedef} - {struct member} - {union member} {
      expandtype [$type type] done
    }
    {struct} - {union} {
      foreach e [$type members] {
        expandtype $e done
      }
    }
  }

  if {[dict exists $done $name]} return
  dict set done $name $type
}


# --- mangle ---------------------------------------------------------------
#
#
proc mangle { type } {
  switch -- [$type kind] {
    enum - struct - union - typedef {
      return [cname [$type fullname]]
    }

    array - sequence {
      if {[catch {$type length} l]} {
        return dyn[$type kind]_[mangle [$type type]]
      } else {
        return [$type kind]${l}_[mangle [$type type]]
      }
    }

    string {
      if {[catch {$type length} l]} {
        return dynstring
      } else {
        return string${l}
      }
    }

    native {
      return [cname [$type kind][$type fullname]]
    }

    {struct member} - {union member} - {const} {
      return [mangle [$type type]]
    }

    default {
      return [cname [$type kind]]
    }
  }

  error "reached unreachable"
}
