<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Aug  1 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <errno.h>
#include <stdio.h>

#include "posterLib.h"

#include "<"$comp">_control_task.h"
#include "serialize.h"


/* --- local data ---------------------------------------------------------- */

static POSTER_ID metadataid;


/* <"[--- genom_metadata_${comp}_init -----------------------------------]"> */

int
genom_metadata_<"$comp">_init(void)
{
  genom_metadata_component meta;
  char name[H2_DEV_MAX_NAME];
  size_t l, m;
  STATUS s;
  char *data;
  int n;

  /* init data */
  genom_tinit_genom_metadata_component(&meta);
  genom_sequence_reserve(&meta.reports,
    <"[llength [[$component event] members]]">);
  m = sizeof(meta.reports._buffer->text);
<'set i 0; foreach m [[$component event] members] {'>
  meta.reports._buffer[<"$i">].id = <"[$m declarator]">;
  strncpy(meta.reports._buffer[<"$i">].text, "<"[$m name]">", m);
  meta.reports._buffer[<"$i">].text[m-1] = 0;
<'  incr i }'>
  meta.reports._length = <"[llength [[$component event] members]]">;

  genom_sequence_reserve(&meta.services, <"[llength [$component services]]">);
  l = sizeof(meta.services._buffer->name);
  m = sizeof(meta.services._buffer->digest);
<'set i 0; foreach s [$component services] {'>
  meta.services._buffer[<"$i">].rqstid = <"$COMP">_<"[$s name]">_RQSTID;
  strncpy(meta.services._buffer[<"$i">].name, "<"[$s name]">", l);
  meta.services._buffer[<"$i">].name[l-1] = 0;
  strncpy(meta.services._buffer[<"$i">].digest, "<"[$s digest]">", m);
  meta.services._buffer[<"$i">].digest[m-1] = 0;
<'  incr i }'>
  meta.services._length = <"[llength [$component services]]">;

  /* create poster */
  n = snprintf(name, sizeof(name), "%s/metadata", genom_instance);
  if (n <= 0 || n >= sizeof(name)) {
    genom_log_warn(0, "port %s name too long", name);
    genom_log_warn(0, "port name length limited to %d characters",
                   H2_DEV_MAX_NAME);
    genom_tfini_genom_metadata_component(&meta);
    return ENAMETOOLONG;
  }
  l = genom_serialen_genom_metadata_component(&meta);
  s = posterCreate(name, (int)l, &metadataid);
  if (s == ERROR) {
    genom_log_warn(1, "cannot create metadata port");
    genom_tfini_genom_metadata_component(&meta);
    return -1;
  }

  s = posterTake(metadataid, POSTER_WRITE);
  if (s != OK) {
    genom_log_warn(1, "cannot access metadata port");
    genom_tfini_genom_metadata_component(&meta);
    return -1;
  }
  data = posterAddr(metadataid);
  genom_serialize_genom_metadata_component(&data, &meta);
  posterGive(metadataid);

  genom_tfini_genom_metadata_component(&meta);
  genom_log_debug("initialized metadata port");
  return 0;
}


/* <"[--- genom_metadata_${comp}_fini -----------------------------------]"> */

void
genom_metadata_<"$comp">_fini(void)
{
  if (metadataid) posterDelete(metadataid);
  metadataid = NULL;
}


/* <"[--- genom_metadata_${comp}_fetch -------------------------------------]"> */

int
genom_metadata_<"$comp">_fetch(const char *comp, genom_metadata_component *meta)
{
  char name[H2_DEV_MAX_NAME];
  POSTER_ID id;
  char *data;
  STATUS s;
  ssize_t m;
  int n;


  n = snprintf(name, sizeof(name), "%s/metadata", comp);
  if (n <= 0 || n >= sizeof(name)) {
    genom_log_warn(0, "port %s name too long", name);
    genom_log_warn(0, "port name length limited to %d characters",
                   H2_DEV_MAX_NAME);
    return ENAMETOOLONG;
  }

  if (posterFind(name, &id) != OK) {
    genom_log_warn(1, "no such port %s", name);
    return ENOENT;
  }

  s = posterTake(id, POSTER_READ);
  if (s != OK) {
    genom_log_warn(1, "cannot access metadata port %s", name);
    return ENOENT;
  }

  data = posterAddr(id);
  m = -1;
  s = genom_deserialize_genom_metadata_component(&data, &m, meta);
  if (s) {
    genom_log_warn(1, "cannot read metadata port %s contents", name);
    return ENOENT;
  }
  posterGive(id);

  return 0;
}
