<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Fri Mar  9 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "h2devLib.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_portlib.h"
#include "serialize.h"

/* --- local data ---------------------------------------------------------- */

enum {
  <"$comp">_PORTLIB_MIN_INDEX = 0,
<'foreach p [$component ports] {'>
  <"$comp">_<"[$p name]">_INDEX,
<'}'>
  <"$comp">_PORTLIB_MAX_INDEX
};

/* --- state port ---------------------------------------------------------- */

int
genom_state_<"$comp">_init(struct genom_<"$comp">_cntrlstr *cids)
{
  <"[[$component event] declarator]"> s;

  /* automatic versioning - nice idea borrowed from ROS */
  strncpy(cids->ports.genom_state.h.buffer.digest, "<"[$component digest]">",
          sizeof(cids->ports.genom_state.h.buffer.digest));
  strncpy(cids->ports.genom_state.h.buffer.date, "<"[clock format [clock sec]]">",
          sizeof(cids->ports.genom_state.h.buffer.date));
  strncpy(cids->ports.genom_state.h.buffer.version, "<"[$component version]">",
          sizeof(cids->ports.genom_state.h.buffer.version));

  s = genom_<"$comp">_genom_state_write();
  if (s) {
    genom_log_warn(0, "cannot initialize state port: %s",
                   genom_<"$comp">_evname(s));
    return 1;
  }

  genom_log_debug("initialized outport genom_state");
  return 0;
}

int
genom_state_<"$comp">_update(struct genom_<"$comp">_cntrlstr *cids)
{
  return 0;
}



<'foreach p [$component ports] {'>

/* <"[--- genom_${comp}_[$p name]_get_ph --------------------------------]"> */

static __inline__ struct genom_<"$comp">_<"[$p name]">_ph *
genom_<"$comp">_<"[$p name]">_get_ph(
<'  if {[$p kind] == "multiple"} {'>
  const char *id,
<'  }'>
  struct genom_<"$comp">_<"[$p name]">_port *p)
{
<'  if {[$p kind] == "multiple"} {'>
  int n;
  for(n = 0; n < p->nh; n++)
    if (p->h[n] && !strcmp(p->h[n]->name, id))
      return p->h[n];
  return NULL;
<'  } else {'>
  return &p->h;
<'  }'>
}


/* <"[--- genom_${comp}_[$p name]_data ----------------------------------]"> */

<"[[$p datatype] argument reference]">
genom_<"$comp">_<"[$p name]">_data(
<'  if {[$p kind] == "multiple"} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return NULL;
  return <"[[$p datatype] pass reference {ph->buffer}]">;
}


/* <"[--- genom_${comp}_[$p name]_open ----------------------------------]"> */

uint32_t
genom_<"$comp">_<"[$p name]">_open(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
  char name[H2_DEV_MAX_NAME];

<'  if {"out" in [$p dir]} {'>
  STATUS s;
  size_t l;
<'  }'>
  int n;

<'  if {"multiple" in [$p kind]} {'>
  n = snprintf(name, sizeof(name), "%s/<"[$p name]">/%s", genom_instance, id);
<'  } else {'>
  n = snprintf(name, sizeof(name), "%s/<"[$p name]">", genom_instance);
<'  }'>
  if (n <= 0 || n >= sizeof(name)) {
    genom_log_warn(0, "port <"[$p name]"> name too long");
    genom_log_warn(0, "port name length limited to %zu characters",
                   H2_DEV_MAX_NAME - strlen(genom_instance) - 2);
    return ENAMETOOLONG;
  }

<'  if {"multiple" in [$p kind]} {'>
  for(n = 0; n < p->nh; n++)
    if (p->h[n] && !strcmp(p->h[n]->name, id)) return EEXIST;
  for(n = 0; n < p->nh; n++)
    if (!p->h[n]) break;
  if (n >= p->nh) {
    struct genom_<"$comp">_<"[$p name]">_ph **pph;
    pph = realloc(p->h, (p->nh+1)*sizeof(*pph));
    if (!pph) return ENOMEM;
    p->h = pph;
    n = p->nh++;
  }
  p->h[n] = malloc(sizeof(*ph));
  if (!p->h[n]) return ENOMEM;

  ph = p->h[n];
  genom_tinit_<"[mangle [$p datatype]]">(
    <"[[$p datatype] pass reference ph->buffer]">);
  strncpy(ph->name, id, sizeof(ph->name));
<'  } else {'>
  n = 0;
  ph = &p->h;
<'  }'>
  ph->id = NULL;

<'  if {"out" in [$p dir]} {'>
  l = genom_serialen_<"[mangle [$p datatype]]">(
    <"[[$p datatype] pass value ph->buffer]">);

  s = posterCreate(name, (int)l, &ph->id);
  if (s == ERROR) {
    genom_log_warn(1, "cannot create outport %s", name);
<'    if {"multiple" in [$p kind]} {'>
    free(ph);
    p->h[n] = NULL;
<'    }'>
    return -1;
  }
  ph->size = l;
  genom_log_info("created outport %s", name);
<'  } else {'>
  genom_log_info("created inport %s", name);
<'  }'>
  return 0;
}


/* <"[--- genom_${comp}_[$p name]_delete --------------------------------]"> */

uint32_t
genom_<"$comp">_<"[$p name]">_close(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return ENOENT;

<'  if {"out" in [$p dir]} {'>
  posterDelete(ph->id);
<'    if {"multiple" in [$p kind]} {'>
  genom_log_info("destroyed outport <"[$p name]">/%s", ph->name);
<'    } else {'>
  genom_log_info("destroyed outport <"[$p name]">");
<'    }'>
<'  } else {'>
<'    if {"multiple" in [$p kind]} {'>
  genom_log_info("destroyed inport <"[$p name]">/%s", ph->name);
<'    } else {'>
  genom_log_info("destroyed inport <"[$p name]">");
<'    }'>
<'  }'>
  ph->id = NULL;
  return 0;
}

void
genom_<"$comp">_<"[$p name]">_delete(void)
{
<'  if {"multiple" in [$p kind]} {'>
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  int n;
  for(n = 0; n < p->nh; n++)
    if (p->h[n]) {
      genom_<"$comp">_<"[$p name]">_close(p->h[n]->name);
      p->h[n] = NULL;
    }
<'  } else {'>
  genom_<"$comp">_<"[$p name]">_close();
<'  }'>
}

<'}'>

<'foreach p [$component ports in] {'>

/* <"[--- genom_${comp}_[$p name]_connect -------------------------------]"> */

<"[[$component event] declarator]">
genom_<"$comp">_<"[$p name]">_connect(
<'  if {"multiple" in [$p kind]} {'>
  const char *id,
<'  }'>
  const char *name, struct genom_<"$comp">_cntrlstr *cids)
{
  struct genom_<"$comp">_<"[$p name]">_port *p = &cids->ports.<"[$p name]">;
  struct genom_<"$comp">_<"[$p name]">_ph *ph;
  POSTER_ID pid;

<'  if {[$p kind] == "multiple"} {'>
  uint32_t s = genom_<"$comp">_<"[$p name]">_open(id);
  if (s && (s != EEXIST))
    return <"[[[$component event] member no_such_inport] declarator]">;
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  
  if (!ph)
    return <"[[[$component event] member no_such_inport] declarator]">;

  if (posterFind(name, &pid) != OK) {
    genom_log_warn(1, "no such port %s", name);
    return <"[[[$component event] member no_such_outport] declarator]">;
  }
  if (posterTake(pid, POSTER_READ) != OK) {
    genom_log_warn(1, "cannot connect port %s", name);
    return <"[[[$component event] member port_io] declarator]">;
  }
  posterGive(pid);

  ph->id = pid;
  return <"[[[$component event] member ok] declarator]">;
}


/* <"[--- genom_${comp}_[$p name]_read ----------------------------------]"> */

uint32_t
genom_<"$comp">_<"[$p name]">_read(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;
  STATUS s;
  char *b;
  ssize_t max;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return ENOENT;

  s = posterTake(ph->id, POSTER_READ);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot access inport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot access inport <"[$p name]">");
<'  }'>
    return <"[[[$component event] member port_io] declarator]">;
  }

  b = posterAddr(ph->id);
  max = -1;
  s = genom_deserialize_<"[mangle [$p datatype]]">(
    &b, &max, <"[[$p datatype] pass value ph->buffer]">);
  posterGive(ph->id);
  if (s) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(0, "cannot read inport <"[$p name]">/%s contents", ph->name);
<'  } else {'>
    genom_log_warn(0, "cannot read inport <"[$p name]"> contents");
<'  }'>
    return <"[[[$component event] member serialization] declarator]">;
  }

  return 0;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- genom_${comp}_[$p name]_write ---------------------------------]"> */

uint32_t
genom_<"$comp">_<"[$p name]">_write(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;
  STATUS s;
  char *b;
  size_t l;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return ENOENT;

  l = genom_serialen_<"[mangle [$p datatype]]">(
    <"[[$p datatype] pass value ph->buffer]">);
  if (l > ph->size) {
    s = posterIoctl(ph->id, FIO_RESIZE, &l);
    if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]">/%s to %zu bytes",
                     ph->name, l);
<'  } else {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]"> to %zu bytes", l);
<'  }'>
      return <"[[[$component event] member serialization] declarator]">;
    }
    ph->size = l;
  }

  s = posterTake(ph->id, POSTER_WRITE);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot access outport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot access outport <"[$p name]">");
<'  }'>
    return <"[[[$component event] member port_io] declarator]">;
  }

  b = posterAddr(ph->id);
  genom_serialize_<"[mangle [$p datatype]]">(
    &b, <"[[$p datatype] pass value ph->buffer]">);
  s = posterGive(ph->id);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot flush outport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot flush outport <"[$p name]">");
<'  }'>
    return <"[[[$component event] member port_io] declarator]">;
  }

  if (l < ph->size) {
    s = posterIoctl(ph->id, FIO_RESIZE, &l);
    if (s != OK)
<'  if {"multiple" in [$p kind]} {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]">/%s to %zu bytes",
                     ph->name, l);
<'  } else {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]"> to %zu bytes", l);
<'  }'>
    else
      ph->size = l;
  }

  return 0;
}

<'}'>

/* --- end of file --------------------------------------------------------- */
